#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#ifdef FFTW
#include <fftw3.h>
#include <omp.h>
#endif

#ifdef clFFT
#include <clFFT.h>
#define CHECK_ERROR(actual, msg) \
    if(actual != CL_SUCCESS) \
    { \
        printf("Location : %d  ", __LINE__); \
		printf(msg); \
		printf("\n");\
		exit(0);\
    }
#endif

#define PRINT_RESULT

const int N_THREADS = 1;
double dt( char ); 

int main( int argc, char* argv[]){
	if(argc < 3){
		printf("wrong parameters\n");
		exit(1);
	} 
	
	int N0 = atoi((const char *)argv[1]);
	int N1 = atoi((const char *)argv[2]);
	printf("Performing fft on an two dimensional array of size N0 x N1 : %d x %d\n", N0,N1);
	double FLOP =2.5*(1e-9)*N0*N1*log(N0*N1)/log(2);

	double total_time=0;
	
	
#ifdef FFTW
	double execute_time = 0.0;
	fftw_plan p;	
	fftw_complex *in;	
	int buffer_size = sizeof(fftw_complex)*N1*(N0/2+1);
	in = (fftw_complex*) fftw_malloc(buffer_size);
#ifndef INPLACE
	double* out = (double*) fftw_malloc(sizeof(double)*N1*N0);
#endif

	for (int i=0; i< N1*(N0/2+1); i++)
    {
	   in[i][0]=(double)i;
	   in[i][1]=(double)1.5*i;
	}
	
#ifdef THREADED
	fftw_init_threads();
	fftw_plan_with_nthreads(N_THREADS);
	printf("Multiple threads: %d threads \n",N_THREADS);
#endif
	
	dt('i');	
	
#ifdef INPLACE
#ifdef FASTPLAN 
	 p = fftw_plan_dft_c2r_2d(N0, N1, in, (double*)in, FFTW_ESTIMATE);
#else
	 p = fftw_plan_dft_c2r_2d(N0, N1, in, (double*)in, FFTW_MEASURE);
#endif
#else
#ifdef FASTPLAN 
	 p = fftw_plan_dft_c2r_2d(N0, N1, in, out, FFTW_ESTIMATE);
#else
	 p = fftw_plan_dft_c2r_2d(N0, N1, in, out, FFTW_MEASURE);
#endif
#endif

	execute_time -= dt('r');
	fftw_execute(p);
	execute_time += dt('r');
	printf("Execute time: %lf \n",execute_time);
	printf("FLOPS = %lf \n", FLOP/execute_time);

#ifdef PRINT_RESULT	
	for (int j=0; j< N1; j++)
    {
		for (int i=0; i< N0; i++)
		{
			#ifdef INPLACE
				printf("%lf \n", ((double*)in)[j*(N0+2)+i]);							
			#else
				printf("%lf \n",out[j*N0+i]);
			#endif
		}
	}	
#endif

#ifdef THREADED
	 fftw_cleanup_threads();
#endif

    fftw_free(in); 
#ifndef INPLACE
	fftw_free(out);
#endif
	fftw_destroy_plan(p);
	
	
#elif defined clFFT
	cl_int err;
	double kernel_time=0;
    cl_platform_id platform = 0;
    cl_device_id device = 0;
    cl_context_properties props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
    cl_context ctx = 0;
    cl_command_queue queue = 0;
    cl_mem bufIn;
    double *In;
    cl_event event = NULL;
    char platform_name[128];
    char device_name[128];
	size_t clLengths[2] = {N0,N1};
	clfftDim dim = CLFFT_2D;
	clfftPlanHandle planHandle;
	
	size_t buffer_size  =  (2 + N0)* N1* sizeof(*In);
    In = (double*)malloc(buffer_size);
#ifndef INPLACE
	cl_mem bufOut;
#endif
    /* print input array */	
	for (int i=0; i< N1*(N0/2+1); i++)
    {
	   In[2*i]=  (double)i;
	   In[2*i+1]=(double)1.5*i;
	}
	
	dt('i');
    /* Setup OpenCL environment. */
    err = clGetPlatformIDs( 1, &platform, NULL );

    size_t ret_param_size = 0;
    err |= clGetPlatformInfo(platform, CL_PLATFORM_NAME,sizeof(platform_name), platform_name,&ret_param_size);
    printf("Platform found: %s\n", platform_name);

    err |= clGetDeviceIDs( platform, CL_DEVICE_TYPE_DEFAULT, 1, &device, NULL );

    err |= clGetDeviceInfo(device, CL_DEVICE_NAME,sizeof(device_name), device_name,&ret_param_size);
	
	CHECK_ERROR(err,"Found devices failed!");
    printf("Device found on the above platform: %s\n", device_name);

    props[1] = (cl_context_properties)platform;
    ctx = clCreateContext( props, 1, &device, NULL, NULL, &err );
    queue = clCreateCommandQueue( ctx, device, CL_QUEUE_PROFILING_ENABLE, &err );
	CHECK_ERROR(err,"creat commandqueue failed!");
	
	/* Setup clFFT. */
    clfftSetupData fftSetup;
    err = clfftInitSetupData(&fftSetup);
    err |= clfftSetup(&fftSetup);
    CHECK_ERROR(err,"clfftSetup failed!");

    /* Prepare OpenCL memory objects and place data inside them. */
    bufIn = clCreateBuffer( ctx, CL_MEM_READ_WRITE, buffer_size, NULL, &err );	
#ifndef INPLACE
	bufOut = clCreateBuffer( ctx, CL_MEM_READ_ONLY, buffer_size, NULL, &err );	
#endif
    CHECK_ERROR(err,"Create Buffer failed!");

    err = clEnqueueWriteBuffer( queue, bufIn, CL_TRUE, 0, buffer_size, In, 0, NULL, NULL );
    CHECK_ERROR(err,"Write Buffer failed!");
	
    /* Create a default plan for a complex FFT. */
    err = clfftCreateDefaultPlan(&planHandle, ctx, dim, clLengths);
    CHECK_ERROR(err,"Create plan failed!");
	
    /* Set plan parameters. */
    err = clfftSetPlanPrecision(planHandle, CLFFT_DOUBLE);
    err |= clfftSetLayout(planHandle, CLFFT_HERMITIAN_INTERLEAVED, CLFFT_REAL);
#ifdef INPLACE
    err |= clfftSetResultLocation(planHandle, CLFFT_INPLACE);
#else
	err |= clfftSetResultLocation(planHandle, CLFFT_OUTOFPLACE);
#endif

    size_t clInStrides[2]={1,N0/2+1};
	size_t clOutStrides[2]={1,N0+2};
	
    err |= clfftSetPlanDistance(planHandle, N1*(N0/2+1),(N0+2)*N1);
    err |= clfftSetPlanInStride(planHandle, dim, clInStrides);
    err |= clfftSetPlanOutStride(planHandle, dim, clOutStrides);

    /* Bake the plan. */
    err |= clfftBakePlan(planHandle, 1, &queue, NULL, NULL);

    /* Execute the plan. */
#ifdef INPLACE
    err |= clfftEnqueueTransform(planHandle, CLFFT_BACKWARD, 1, &queue, 0, NULL, &event, &bufIn, NULL, NULL);
#else
	err |= clfftEnqueueTransform(planHandle, CLFFT_BACKWARD, 1, &queue, 0, NULL, &event, &bufIn, &bufOut, NULL);
#endif

    /* Wait for calculations to be finished. */
    err |= clFinish(queue);
	CHECK_ERROR(err,"run failed!");
	
	cl_ulong start=0; 
	cl_ulong stop=0;
	err = clGetEventProfilingInfo(event,CL_PROFILING_COMMAND_START, sizeof(cl_ulong),&start, NULL);
	err |= clGetEventProfilingInfo(event,CL_PROFILING_COMMAND_END, sizeof(cl_ulong),&stop, NULL);
	CHECK_ERROR(err,"Get kernel time failed!");
	
	kernel_time = 1.0*(stop -start)/1e9;
	printf("Kernel time = %lf \n", kernel_time);
	printf("FLOPS = %lf \n", FLOP/kernel_time);
	printf("BS = %lf \n", 10e-9*buffer_size/kernel_time);

    /* Fetch results of calculations. */
#ifdef INPLACE
    err = clEnqueueReadBuffer( queue, bufIn, CL_TRUE, 0, buffer_size, In, 0, NULL, NULL );
#else
	err = clEnqueueReadBuffer( queue, bufOut, CL_TRUE, 0, buffer_size,In, 0, NULL, NULL );
#endif
	
	CHECK_ERROR(err,"Read Buffer failed!");

#ifdef PRINT_RESULT
    /* print output array */		
	for (int j=0; j< N1; j++)
    {
		for (int i=0; i< N0; i++)
		{
			#ifdef INPLACE
				printf("%lf \n", In[j*(N0+2)+i]);							
			#else
				printf("%lf \n", Out[j*(N0+2)+i]);
			#endif
		}
	}	
	
#endif
    /* Release OpenCL memory objects. */
    clReleaseMemObject( bufIn );
#ifndef INPLACE
	clReleaseMemObject( bufOut );
#endif

    free(In);
    /* Release the plan. */
    err = clfftDestroyPlan( &planHandle );
    /* Release clFFT library. */
    clfftTeardown( );
    /* Release OpenCL working objects. */
    clReleaseCommandQueue( queue );
    clReleaseContext( ctx );
	
#endif	
	total_time = dt('e');
	printf("Total time: %lf \n\n",total_time);

}